#!/usr/bin/perl

use strict;
use warnings;

my %loc = (); # key=symbol, value=location (or empty string)
my $seenDups = 0;

sub acceptDupSym($) {
  # return true for all symbols that shall be accepted if duplicated
  my ($sym) = @_;
  return $sym =~ /^(ARB_main|aw_never_called_main)\(/o;
}

sub isGlobalSymFlag($) {
  # return true for all flags that shall be checked for duplicates
  my ($f) = @_;
  if (($f ge 'A' and $f le 'D') or ($f eq 'G') or ($f ge 'R' and $f le 'T')) {
    return 1;
  }
  return 0;
}

sub addSym($$) {
  my ($sym,$loc) = @_;
  my $firstLoc = $loc{$sym};
  if (defined $firstLoc) {
    if ($firstLoc eq '') { # if first seen location was missing => use this location
      $loc{$sym} = $loc;
    }
    elsif ($loc eq '') {
      ; # ignore missing locations after one explicit location was seen
    }
    elsif ($firstLoc ne $loc) {
      if (not acceptDupSym($sym)) {
        $seenDups++;

        print "$firstLoc: Warning: first occurrence of $sym\n";
        print "$loc: Warning: duplicated occurrence of $sym\n";
      }
    }
  }
  else {
    $loc{$sym} = $loc;
  }
}

my $reg_sym = qr/^[0-9a-f]*\s+([A-Z])\s(.*)$/i;
my $reg_loc = qr/^(.*)\s([^\s]+:[0-9]+)$/i;

sub scanSymdef($) {
  my ($line) = @_;
  if ($line ne '') {
    if ($line =~ $reg_sym) {
      my ($flag,$rest) = ($1,$2);
      if (isGlobalSymFlag($flag)) {
        if ($rest =~ $reg_loc) {
          my ($proto,$loc) = ($1,$2);
          addSym($proto,$loc);
        }
        else { # assume whole rest is prototype
          addSym($rest,'');
        }
      }
    }
    else {
      if ($line =~ /^[^\s]+:$/o) { # accept start of single objects
        ;
      }
      else {
        die "cannot parse line '$line'\n";
      }
    }
  }
}

sub scanSymfile($) {
  my ($symfile) = @_;
  open(IN,'<'.$symfile) || die "failed to open '$symfile' (Reason: $!)";

  my $lineno = 0;
  eval {
    my $line;
    while (defined($line=<IN>)) {
      chomp($line);
      $lineno++;
      scanSymdef($line);
    }
  };
  if ($@) {
    die "$symfile:$lineno: $@";
  }
  close(IN);
}

# --------------------------------------------------------------------------------

sub main() {
  my $args = scalar(@ARGV);
  if ($args<1) {
    die "Usage: dupsymwarn.pl nm-output [nm-output]+\n".
      "Collects and warns about duplicated global symbols\n";
  }

  foreach my $symfile (@ARGV) {
    scanSymfile($symfile)
  }

  if ($seenDups>0) {
    print "Encountered $seenDups duplicated global symbols in arb-codebase\n";
    print "Notes:\n";
    print " * this is checked in TESTED modules and libraries only (i.e. it may miss some dups)\n";
    print " * this is unwanted => please fix it!\n";
  }
}
main();
